;; Generated package description from vecdb.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "vecdb" "0.2.2.0.20251011.220856" "An interface to vector databases for embeddings" '((emacs "29.1") (plz "0.8") (pg "0.56")) :commit "57dfaf751f9d71e921cdbeaa34b9aafa3c502ba3" :authors '(("Andrew Hyatt" . "ahyatt@gmail.com")) :maintainer '("Andrew Hyatt" . "ahyatt@gmail.com") :url "https://github.com/ahyatt/vecdb")
